from dataclasses import dataclass

from rest_framework.exceptions import ValidationError

from drf_service_layer.services import Service
from tests.test_app.models import Product


@dataclass
class PriceDTO:
    currency: str


class ProductService(Service):
    @staticmethod
    def filter_currency_param(currency_param: str) -> str:
        if currency_param not in ["WON", "EUR"]:
            raise ValidationError("This currency is not supported.")
        return currency_param

    def adjust_price(self, instance: Product) -> float:
        self.dto: PriceDTO

        exchange_rate = {
            "WON": 1200,
            "EUR": 0.85,
        }

        return instance.price * exchange_rate[self.dto.currency]



from rest_framework import generics
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from drf_service_layer import views
from tests.test_app.models import Product
from tests.test_app.serializers import DRFProductSerializer, ProductSerializer
from tests.test_app.services import PriceDTO, ProductService


class ProductRetrieveView(views.RetrieveAPIView):
    queryset = Product.objects.all()
    serializer_class = ProductSerializer
    service_class = ProductService

    @property
    def dto(self):
        currency_param = self.request.query_params.get("currency", "WON")
        currency = self.pre_service.filter_currency_param(currency_param)

        return PriceDTO(
            currency=currency,
        )


class ProductListView(views.ListAPIView):
    queryset = Product.objects.all()
    serializer_class = ProductSerializer
    service_class = ProductService

    @property
    def dto(self):
        currency_param = self.request.query_params.get("currency", "WON")
        currency = self.pre_service.filter_currency_param(currency_param)

        return PriceDTO(
            currency=currency,
        )


# Original way of implementation


class DRFProductRetrieveView(generics.RetrieveAPIView):
    queryset = Product.objects.all()
    serializer_class = DRFProductSerializer

    def retrieve(self, request, *args, **kwargs):
        currency_param = self.request.query_params.get("currency", "WON")
        if currency_param not in ["WON", "EUR"]:
            raise ValidationError("This currency is not supported.")

        currency = currency_param

        exchange_rate = {
            "WON": 1200,
            "EUR": 0.85,
        }

        instance = self.get_object()
        instance.price = instance.price * exchange_rate[currency]

        serializer = self.get_serializer(instance)
        return Response(serializer.data)


# Error cases


class ErrorCaseView(views.RetrieveAPIView):
    queryset = Product.objects.all()
    serializer_class = ProductSerializer



import pytest

from tests.test_app.models import Product


@pytest.fixture(autouse=True)
def instance(db):
    yield Product.objects.create(
        name="Long-sleeve T-Shirts",
        price=15.5,
        category="C",
        description="2021 F/W",
    )



import pytest
from django.shortcuts import resolve_url
from rest_framework.test import APIClient
from schema import And, Schema

from drf_service_layer.services import Service
from tests.test_app.serializers import ErrorCaseSerializer
from tests.test_app.services import ProductService


class TestServiceAPIView:
    @pytest.mark.parametrize(
        "currency, adjusted_price",
        [
            ("WON", 15.5 * 1200),
            ("EUR", 15.5 * 0.85),
        ],
    )
    def test_product_retrieve(self, currency, adjusted_price):
        client = APIClient()
        url = resolve_url("product_retrieve_service", pk=1)

        query_param = {"currency": currency}
        response = client.get(url, query_param)

        schema = Schema(
            {
                "id": 1,
                "name": str,
                "price": adjusted_price,
                "category": And(str, lambda s: len(s) == 1),
                "description": str,
            }
        )

        assert response.status_code == 200
        assert schema.is_valid(response.json())

    @pytest.mark.parametrize(
        "currency, adjusted_price",
        [
            ("WON", 15.5 * 1200),
            ("EUR", 15.5 * 0.85),
        ],
    )
    def test_product_list(self, currency, adjusted_price):
        client = APIClient()
        url = resolve_url("product_list_service")

        query_param = {"currency": currency}
        response = client.get(url, query_param)

        schema = Schema(
            [
                {
                    "id": int,
                    "name": str,
                    "price": adjusted_price,  # return 1 instance
                    "category": And(str, lambda s: len(s) == 1),
                    "description": str,
                }
            ]
        )

        assert response.status_code == 200
        assert schema.is_valid(response.json())


# Original way of implementation


class TestDRFAPIView:
    @pytest.mark.parametrize(
        "currency, adjusted_price",
        [
            ("WON", 15.5 * 1200),
            ("EUR", 15.5 * 0.85),
        ],
    )
    def test_product_retrieve(self, currency, adjusted_price):
        client = APIClient()
        url = resolve_url("product_retrieve_original", pk=1)

        query_param = {"currency": currency}
        response = client.get(url, query_param)

        schema = Schema(
            {
                "id": 1,
                "name": str,
                "price": adjusted_price,
                "category": And(str, lambda s: len(s) == 1),
                "description": str,
            }
        )

        schema.validate(response.json())

        assert response.status_code == 200
        assert schema.is_valid(response.json())


# Error cases


class TestViewErrors:
    def test_service_class_not_defined(self):
        client = APIClient()
        url = resolve_url("product_retrieve_errors", pk=1)

        with pytest.raises(
            AssertionError,
            match=r".+ should either include a `service_class` attribute.",
        ):
            client.get(url)


class TestDecoratorErrors:
    def test_serializer_context(self, instance):
        with pytest.raises(
            AssertionError, match=r".+ should retrieve service context from view."
        ):
            ErrorCaseSerializer(instance)

    def test_service_class_match(self, instance):
        context = {"service": Service(None)}
        with pytest.raises(
            AssertionError,
            match=r".+ - The service injected from view and the service declared in the decorator are not the same..+",
        ):
            ErrorCaseSerializer(instance, context=context)

    def test_dto_match(self, instance):
        context = {"service": ProductService(None)}
        with pytest.raises(
            AssertionError,
            match=r".+ - The dto injected from view and the dto declared in the decorator are not matched..+",
        ):
            ErrorCaseSerializer(instance, context=context)



from django.contrib.auth.models import AbstractUser
from django.db import models


class User(AbstractUser):
    pass


CATEGORY_CHOICES = (("C", "Clothes"), ("F", "Food"), ("B", "Beverage"))

ORDER_STATUS_CHOICES = (
    ("P", "Pending"),
    ("R", "Ready for delivery"),
    ("S", "Shipped"),
    ("R", "Received"),
)


class Product(models.Model):
    name = models.CharField(max_length=128)
    price = models.FloatField()
    category = models.CharField(choices=CATEGORY_CHOICES, max_length=1)
    description = models.TextField()


class Order(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    ordered_at = models.DateTimeField(auto_now_add=True)
    canceled_at = models.DateTimeField()
    is_paid = models.BooleanField(default=False)
    is_canceled = models.BooleanField(default=False)
    status = models.CharField(choices=CATEGORY_CHOICES, max_length=1)


class OrderDetail(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    product = models.ForeignKey(Product, on_delete=models.CASCADE)
    quantity = models.PositiveIntegerField(default=1)
    order = models.ForeignKey(Order, on_delete=models.CASCADE, null=True)



from rest_framework import mixins
from rest_framework.viewsets import ViewSetMixin

from .views import GenericServiceAPIView


class GenericViewSet(ViewSetMixin, GenericServiceAPIView):
    """
    Service-layered viewset.
    The GenericViewSet class does not provide any actions by default,
    but does include the base set of generic view behavior, such as
    the `get_object` and `get_queryset` methods.
    """

    pass


class ReadOnlyModelViewSet(
    mixins.RetrieveModelMixin, mixins.ListModelMixin, GenericViewSet
):
    """
    A viewset that provides default `list()` and `retrieve()` actions.
    """

    pass


class ModelViewSet(
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    GenericViewSet,
):
    """
    A viewset that provides default `create()`, `retrieve()`, `update()`,
    `partial_update()`, `destroy()` and `list()` actions.
    """

    pass



from rest_framework import serializers

from drf_service_layer.services import service_layer
from tests.test_app.models import Product
from tests.test_app.services import PriceDTO, ProductService


@service_layer(ProductService, PriceDTO)
class ProductSerializer(serializers.ModelSerializer):
    class Meta:
        model = Product
        fields = "__all__"

    def to_representation(self, instance):
        representation = super().to_representation(instance)
        representation["price"] = self.service.adjust_price(instance)
        return representation


# Original way of implementation


class DRFProductSerializer(serializers.ModelSerializer):
    class Meta:
        model = Product
        fields = "__all__"


# Error cases


@service_layer(ProductService, PriceDTO)
class ErrorCaseSerializer(serializers.ModelSerializer):
    class Meta:
        model = Product
        fields = "__all__"



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['drf_service_layer']

package_data = \
{'': ['*']}

install_requires = \
['django', 'djangorestframework']

extras_require = \
{'dev': ['pytest >=6.2.4,<7.0.0',
         'pytest-django ==4.4.0',
         'pytest-cov >=2.12.0,<3.0.0',
         'flake8 >=3.8.3,<4.0.0',
         'black ==20.8b1',
         'isort >=5.0.6,<6.0.0']}

setup(name='drf-service-layer',
      version='0.0.10',
      description='Simple package which supports service-layered design for Django REST Framework.',
      author=None,
      author_email='qu3vipon <qu3vipon@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )



from typing import Any

from django.contrib.auth import get_user_model

User = get_user_model()


class Service:
    def __init__(self, dto: Any):
        self.dto = dto


def service_layer(service_class, dto_class=None):
    """
    add service layer to Serializer
    """

    def wrapper(cls):
        original_init = cls.__init__

        def __init__(self, *args, **kwargs):
            original_init(self, *args, **kwargs)

            assert (
                self.context.get("service") is not None
            ), f"{self.__class__.__name__} should retrieve service context from view."

            service = self.context["service"]
            assert isinstance(
                service, service_class
            ), f"{self.__class__.__name__} - The service injected from view and the service declared in the decorator are not the same.({service} - {service_class})"

            setattr(self, "service", service)

            dto = service.dto
            if dto_class:
                assert isinstance(
                    dto, dto_class
                ), f"{self.__class__.__name__} - The dto injected from view and the dto declared in the decorator are not matched.({service.dto} - {dto})"

            setattr(self, "dto", dto)

        cls.__init__ = __init__
        return cls

    return wrapper



from rest_framework import generics, mixins


class GenericServiceAPIView(generics.GenericAPIView):
    service_class = None
    service = None
    pre_service = None
    dto = None

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        self.attach_service()

    def attach_service(self, *args, **kwargs):
        assert self.service_class is not None, (
            "'%s' should either include a `service_class` attribute."
            % self.__class__.__name__
        )

        self.pre_service = self.service_class(None)

        try:
            dto = getattr(self, "dto")
        except AttributeError:
            dto = None

        self.service = self.service_class(dto)

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        return {
            "request": self.request,
            "format": self.format_kwarg,
            "view": self,
            "service": self.service,
        }


class CreateAPIView(mixins.CreateModelMixin, GenericServiceAPIView):
    """
    Service-layered view for creating a model instance.
    """

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class ListAPIView(mixins.ListModelMixin, GenericServiceAPIView):
    """
    Service-layered view for listing a queryset.
    """

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)


class RetrieveAPIView(mixins.RetrieveModelMixin, GenericServiceAPIView):
    """
    Service-layered view for retrieving a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)


class DestroyAPIView(mixins.DestroyModelMixin, GenericServiceAPIView):
    """
    Service-layered view for deleting a model instance.
    """

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)


class UpdateAPIView(mixins.UpdateModelMixin, GenericServiceAPIView):
    """
    Service-layered view for updating a model instance.
    """

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class ListCreateAPIView(
    mixins.ListModelMixin, mixins.CreateModelMixin, GenericServiceAPIView
):
    """
    Service-layered view for listing a queryset or creating a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class RetrieveUpdateAPIView(
    mixins.RetrieveModelMixin, mixins.UpdateModelMixin, GenericServiceAPIView
):
    """
    Service-layered view for retrieving, updating a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class RetrieveDestroyAPIView(
    mixins.RetrieveModelMixin, mixins.DestroyModelMixin, GenericServiceAPIView
):
    """
    Service-layered view for retrieving or deleting a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)


class RetrieveUpdateDestroyAPIView(
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    GenericServiceAPIView,
):
    """
    Service-layered view for retrieving, updating or deleting a model instance.
    """

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)
